import inspect
import logging
import sys
import types
from typing import Type, Union

logger = logging.getLogger("tydi.inspectors.method_inspector")


class MethodInspector:
    @staticmethod
    def is_class_method(method: object) -> bool:
        """
        Check if the provided object is a class method.

        :param method: The object to check.
        :return: True if the object is a method or a function, otherwise False.
        """
        return (
            inspect.ismethod(method)
            or inspect.isfunction(method)
            and hasattr(method, "__qualname__")
            and "." in method.__qualname__
        )

    @staticmethod
    def get_name(method: Union[types.MethodType, types.FunctionType]) -> str:
        """
        Get the name of the method or function.

        :param method: The method or function to inspect.
        :return: The name of the method or function.
        """
        return method.__name__

    @staticmethod
    def get_class(method: Union[types.MethodType, types.FunctionType]) -> Type:
        """
        Get the class of the method or function.

        :param method: The method or function to inspect.
        :return: The class that defines the method or function.
        """
        if isinstance(method, property):
            method = method.fget

        if hasattr(method, "__self__"):
            # Bound method
            return method.__self__.__class__

        # Extract the class name from __qualname__
        class_name = method.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]

        # Get the module where the method is defined
        module_name = method.__module__
        module = sys.modules[module_name]

        # Retrieve and return the class object
        return getattr(module, class_name)

    @staticmethod
    def get_class_name(
        method: Union[types.MethodType, types.FunctionType],
    ) -> str:
        """
        Get the class name of the method or function.

        :param method: The method or function to inspect.
        :return: The name of the class that defines the method or function.
        """
        return method.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]



import importlib
import inspect
import logging
from typing import Callable, List

from ..overload import get_overloads

logger = logging.getLogger("tydi.inspectors.module_inspector")


class ModuleInspector:
    @staticmethod
    def is_module_function(func: Callable) -> bool:
        return inspect.isfunction(func) and not inspect.ismethod(func)

    @staticmethod
    def get_overloaded_functions(
        module_name: str, function_name: str
    ) -> List[Callable]:
        """
        Retrieve all overloaded functions with the given name from the specified module.

        Args:
            module_name (str): The name of the module to inspect.
            function_name (str): The name of the function to find overloads for.

        Returns:
            List[Callable]: A list of Callable objects representing the overloaded functions.
        """
        try:
            # Ensure the module is imported and get a reference to it
            module = importlib.import_module(module_name)
        except ImportError:
            raise ImportError(f"Module '{module_name}' could not be imported.")

        # Get the function object
        func = getattr(module, function_name, None)
        if func is None:
            return []

        # Get the overloads
        overloads = get_overloads(func)
        logger.debug(
            f"Found {len(overloads)} overloads for function '{function_name}' in module '{module_name}'."
        )

        # If no overloads are found, it might mean the function isn't overloaded,
        # or we're using a Python version earlier than 3.11
        if not overloads:
            return [func] if hasattr(func, "__overload__") else []

        return list(overloads)



from .class_inspector import ClassInspector
from .method_inspector import MethodInspector
from .module_inspector import ModuleInspector

__all__ = ["ClassInspector", "MethodInspector", "ModuleInspector"]



import asyncio
import logging
from asyncio import iscoroutinefunction
from inspect import signature
from typing import Any, Callable, Dict, List, Tuple, TypeVar, get_type_hints

from beartype.door import is_bearable

from .inspectors import ClassInspector, MethodInspector, ModuleInspector

logger = logging.getLogger("tydi.multi_dispatch")

T = TypeVar("T")


class MultiMethod:
    def __init__(self, func: Callable, is_inspected: bool = False):
        self.is_inspected = is_inspected
        self.func = func
        self.methods: List[Callable] = []

    def register(self, func: Callable):
        self.methods.append(func)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        for method in self.methods:
            if self._match_args(method, args, kwargs):
                logger.debug(
                    f"Matched the args {args} and kwargs {kwargs} with method signature: {get_type_hints(method)}"
                )
                if iscoroutinefunction(method):
                    return asyncio.create_task(method(*args, **kwargs))
                return method(*args, **kwargs)
        raise TypeError(
            f"No matching method found for {self.func} with given arguments"
        )

    def _match_args(
        self, func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> bool:
        sig = signature(func)
        try:
            bound_args = sig.bind(*args, **kwargs)
            type_hints = get_type_hints(func)
            logger.debug(
                f"\n\nMatching the bound arguments {bound_args.arguments} with func type hints: {type_hints}"
            )
            matched_params = set()
            for param_name, arg in bound_args.arguments.items():
                logger.debug(
                    f"Checking argument: {param_name} = {arg} of type {type(arg)}"
                )

                if param_name == "self":
                    matched_params.add(param_name)
                    continue

                if param_name in type_hints:
                    if not is_bearable(arg, type_hints[param_name]):
                        logger.debug(
                            f"Argument {param_name} of type {type(arg)} does not match func type hint {type_hints[param_name]}"
                        )
                        return False
                    else:
                        logger.debug(
                            f"Argument {param_name} of type {type(arg)} matches func type hint {type_hints[param_name]}"
                        )
                        matched_params.add(param_name)
                else:
                    logger.debug(
                        f"Argument {param_name} is not in func type hints, type check failed"
                    )
                    return False

            unmatched_params = set()
            for param in sig.parameters:
                if param not in matched_params:
                    unmatched_params.add(param)
            if unmatched_params:
                logger.debug(
                    f"Unmatched parameters: {unmatched_params}, expected parameters: {sig.parameters.keys()}"
                )
                return False
            return True
        except TypeError:
            return False


class MultiDispatch:
    def __init__(self):
        self.registry: Dict[Callable[..., T], MultiMethod] = dict()

    def _is_registered(self, func: Callable[..., T]) -> bool:
        return func in self.registry

    def _get_overloaded_functions(
        self, func: Callable[..., T]
    ) -> List[Callable[..., T]]:
        overloaded_funcs = []
        if MethodInspector.is_class_method(func):
            overloaded_funcs = ClassInspector.get_overloaded_methods(
                MethodInspector.get_name(func),
                MethodInspector.get_class(func),
            )
        elif ModuleInspector.is_module_function(func):
            overloaded_funcs = ModuleInspector.get_overloaded_functions(
                func.__module__, func.__name__
            )
        return overloaded_funcs

    def _register_overloaded_functions(self, func, overloaded_funcs) -> None:
        for _func in overloaded_funcs:
            self.registry[func].register(_func)
        return

    def register(self, func: Callable[..., T]) -> None:
        if self._is_registered(func):
            return

        self.registry[func] = MultiMethod(func, is_inspected=False)
        return

    def dispatch(self, func: Callable[..., T], *args, **kwargs) -> T:
        multi_method = self.registry[func]
        if not multi_method.is_inspected:
            overloaded_funcs = self._get_overloaded_functions(func)
            self._register_overloaded_functions(func, overloaded_funcs)
            self.registry[func].is_inspected = True
        return multi_method(*args, **kwargs)


dispatcher = MultiDispatch()

__all__ = ["dispatcher"]



from functools import wraps

from .multi_dispatch import dispatcher


def dispatch(func):
    dispatcher.register(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        return dispatcher.dispatch(func, *args, **kwargs)

    return wrapper



import sys

if sys.version_info >= (3, 11):  # pragma: specific no cover 3.7 3.8 3.9 3.10
    from typing import get_overloads, overload
else:  # pragma: specific no cover 3.11
    from typing_extensions import get_overloads, overload


__all__ = ["overload", "get_overloads"]



import inspect
import logging
from types import FunctionType
from typing import Callable, Dict, List, Type

from ..overload import get_overloads

logger = logging.getLogger("tydi.inspectors.class_inspector")


class ClassInspector:
    """
    A utility class for inspecting class information, including methods,
    superclasses, and method resolution order (MRO).
    """

    @staticmethod
    def get_class_name(cls: Type) -> str:
        """
        Get the name of the class.

        Args:
            cls (Type): The class type to inspect.

        Returns:
            str: The name of the class.
        """
        return cls.__name__

    @staticmethod
    def get_superclasses(cls: Type) -> tuple[type, ...]:
        """
        Get a list of all superclasses of the given class.

        Args:
            cls (Type): The class type to inspect.

        Returns:
            List[Type]: A list of superclasses of the class.
        """
        return cls.__mro__[1:]  # Skip the class itself

    @staticmethod
    def get_all_methods(cls: Type) -> Dict[str, List[FunctionType]]:
        """
        Get all methods of the class, including inherited methods.

        Args:
            cls (Type): The class type to inspect.

        Returns:
            Dict[str, List[FunctionType]]: A dictionary where keys are method names and values are method functions.
        """
        methods: Dict[str, List[FunctionType]] = {}
        for cls in inspect.getmro(cls):
            for name, func in inspect.getmembers(
                cls,
                predicate=inspect.isfunction,
                # predicate=inspect.ismethod,
            ):
                if name not in methods:
                    methods[name] = []
                methods[name].append(func)
        return methods

    @staticmethod
    def get_method_mro(method_name: str, cls: Type) -> List[Callable]:
        """
        Get the method resolution order (MRO) for a given method name.

        Args:
            method_name (str): The name of the method to inspect.
            cls (Type): The class type to inspect.

        Returns:
            List[Callable]: A list of methods of same name in MRO.
        """
        method_mro = []
        for base in inspect.getmro(cls):
            if method_name in base.__dict__:
                method = base.__dict__[method_name]
                method_mro.append(method)
        return method_mro

    @staticmethod
    def get_overloaded_methods(method_name: str, cls: Type) -> List[Callable]:
        """
        Get all overloaded methods with the given name in the current class.

        Args:
            method_name (str): The name of the method to inspect.
            cls (Type): The class type to inspect.

        Returns:
            List[Callable]: A list of overloaded methods with the given name.
        """
        methods = []
        class_dict = vars(cls)
        if method_name in class_dict:
            method = class_dict[method_name]
            if isinstance(method, classmethod):
                method = method.__func__
            overloads = get_overloads(method)
            if overloads:
                methods.extend(overloads)
            else:
                methods.append(method)
        return methods
